from __future__ import annotations

import os
import typing

from dataclasses import dataclass

import sqlalchemy as sa

from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine.url import URL
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    create_async_engine,
)


from .model import CRUDMixin, Model, ID, Field


@dataclass
class Database:
    """Defines database."""

    dsn: str
    engine: typing.Optional[AsyncEngine]

    driver: str = 'postgresql+asyncpg'

    @staticmethod
    def from_env() -> typing.Optional[Database]:
        """Read definition from env."""
        dsn = URL.create(
            drivername=Database.driver,
            host=os.getenv('DB_HOST'),
            port=os.getenv('DB_PORT'),
            database=os.getenv('DB_NAME'),
            username=os.getenv('DB_USER'),
            password=os.getenv('DB_PASSWORD'),
        )
        debug = os.getenv('DEBUG', False)
        engine = create_async_engine(dsn, future=True, echo=debug)
        return Database(dsn=dsn, engine=engine)


async def get_session(engine: AsyncEngine) -> typing.Optional[AsyncSession]:
    """Get session."""
    new = sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)
    async with engine.begin() as connection:
        async with new(bind=connection) as async_session:
            yield async_session
            await async_session.commit()


__all__ = (
    # Base
    'Database',
    'get_session',
    'Model',
    'ID',
    'Field',

    # Shortcuts
    'sa',
    'AsyncSession',
    'AsyncEngine',
)
