import typing

from sqlmodel import SQLModel, Field
from sqlalchemy import select, update
from sqlalchemy.ext.asyncio import AsyncSession


class CRUDMixin(SQLModel, table=False):
    """CRUD Mixin."""

    @classmethod
    async def item_detail(model, session: AsyncSession, id: int):
        """Detail item."""
        return await session.get(model, id)

    @classmethod
    async def item_all(model, session: AsyncSession):
        """All items."""
        items = await session.execute(select(model))
        return [model(**item[0].dict()) for item in items]

    @classmethod
    async def item_add(
        model,
        session: AsyncSession,
        fields:
        typing.Dict[str, typing.Any]
    ):
        """Add item."""
        fields.pop('id', None)
        new = model(**fields)
        session.add(new)
        await session.commit()
        await session.refresh(new)
        return new

    @classmethod
    async def item_delete(model, session: AsyncSession, id: int):
        """Delete item."""
        removable = await session.get(model, id)
        await session.delete(removable)
        await session.commit()
        await session.close()

    @classmethod
    async def item_update(
        model,
        session: AsyncSession,
        id: int,
        fields: typing.Dict[str, typing.Any],
    ):
        """Update item."""
        await session.execute(update(model).where(model.id == id).values(**fields))
        await session.commit()
        await session.close()


class Model(CRUDMixin, table=False):
    """Model."""

    pass


ID = Field(default=None, primary_key=True)


__all__ = ('CRUDMixin', 'Model', 'ID', 'Field')
